/*
    QuickMSG
    Copyright (C) 2014  Jeroen Vreeken <jeroen@vreeken.net>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/
#define _GNU_SOURCE
#include "message.h"
#include "attachment.h"
#include "settings.h"

#include <libetpan/libetpan.h>

#include <string.h>

void display_mime_parameter(struct mailmime_parameter * param)
{
  printf("%s = %s\n", param->pa_name, param->pa_value);
}

void display_mime_content(struct mailmime_content * content_type)
{
  clistiter * cur;

  printf("type:\n");
  //display_type(content_type->ct_type);
  printf("\n");
  printf("subtype: %s\n", content_type->ct_subtype);
  printf("\n");
  
  for(cur = clist_begin(content_type->ct_parameters) ; cur != NULL ;
    cur = clist_next(cur)) {
    struct mailmime_parameter * param;
    
    param = clist_content(cur);
    display_mime_parameter(param);
    printf("\n");
  }
  printf("\n");
}

void display_field(struct mailimf_field * field)
{
  switch (field->fld_type) {
  case MAILIMF_FIELD_RETURN_PATH:
    printf("Return-Path:\n");
//    display_return(field->fld_data.fld_return_path);
    break;
  case MAILIMF_FIELD_RESENT_DATE:
    printf("Resent-Date:\n");
//    display_orig_date(field->fld_data.fld_orig_date);
    break;
  case MAILIMF_FIELD_RESENT_FROM:
    printf("Resent-From:\n");
//    display_from(field->fld_data.fld_orig_date);
    break;
  case MAILIMF_FIELD_RESENT_SENDER:
    printf("Resent-Sender:\n");
//    display_sender(field->fld_data.fld_resent_sender);
    break;
  case MAILIMF_FIELD_RESENT_TO:
    printf("Resent-To:\n");
//    display_to(field->fld_data.fld_resent_to);
    break;
  case MAILIMF_FIELD_RESENT_CC:
    printf("Resent-Cc:\n");
//    display_from(field->fld_data.fld_resent_cc);
    break;
  case MAILIMF_FIELD_RESENT_BCC:
    printf("Resent-Bcc:\n");
//    display_from(field->fld_data.fld_resent_bcc);
    break;
  case MAILIMF_FIELD_RESENT_MSG_ID:
    printf("Resent-Message-ID:\n");
//    display_message_id(field->fld_data.fld_resent_msg_id);
    break;
  case MAILIMF_FIELD_ORIG_DATE:
    printf("Date:\n");
//    display_orig_date(field->fld_data.fld_orig_date);
    break;
  case MAILIMF_FIELD_FROM:
    printf("From:\n");
//    display_from(field->fld_data.fld_from);
    break;
  case MAILIMF_FIELD_SENDER:
    printf("Sender:\n");
//     display_sender(field->fld_data.fld_sender);
    break;
  case MAILIMF_FIELD_REPLY_TO:
    printf("Reply-To:\n");
//    display_reply_to(field->fld_data.fld_reply_to);
    break;
  case MAILIMF_FIELD_TO:
    printf("To:\n");
//    display_to(field->fld_data.fld_to);
    break;
  case MAILIMF_FIELD_CC:
    printf("Cc:\n");
//    display_cc(field->fld_data.fld_cc);
    break;
  case MAILIMF_FIELD_BCC:
    printf("Bcc:\n");
//    display_bcc(field->fld_data.fld_bcc);
    break;
  case MAILIMF_FIELD_MESSAGE_ID:
    printf("Message-ID:\n");
//    display_message_id(field->fld_data.fld_message_id);
    break;
  case MAILIMF_FIELD_IN_REPLY_TO:
    printf("In-Reply-To:\n");
//    display_in_reply_to(field->fld_data.fld_in_reply_to);
    break;
  case MAILIMF_FIELD_REFERENCES:
    printf("References:\n");
//    display_references(field->fld_data.fld_references_to);
    break;
  case MAILIMF_FIELD_SUBJECT:
    printf("Subject:\n");
//    display_subject(field->fld_data.fld_subject);
    break;
  case MAILIMF_FIELD_COMMENTS:
    printf("Comments:\n");
//    display_comments(field->fld_data.fld_comments);
    break;
  case MAILIMF_FIELD_KEYWORDS:
    printf("Keywords:\n");
//    display_keywords(field->fld_data.fld_keywords);
    break;
  case MAILIMF_FIELD_OPTIONAL_FIELD:
    printf("[optional field]:\n");
//    display_optional_field(field->fld_data.fld_optional_field);
    break;
  }

}

void display_fields(struct mailimf_fields * fields)
{
  clistiter * cur;

  for(cur = clist_begin(fields->fld_list) ; cur != NULL ;
    cur = clist_next(cur)) {
    struct mailimf_field * f;
    
    f = clist_content(cur);
    
    display_field(f);
    printf("\n");
  }
}

int message_parse_from(char *msg, size_t length, char **from)
{
	int r;
	struct mailimf_fields * f;
	size_t current_index = 0;
	clistiter * cur;
	
	*from = NULL;	
	
	r = mailimf_envelope_fields_parse(msg, length, &current_index, &f);
	if (r != MAILIMF_NO_ERROR)
		return -1;

	for(cur = clist_begin(f->fld_list) ; cur != NULL ;
	    cur = clist_next(cur)) {
		struct mailimf_field * ef;
    
		ef = clist_content(cur);
    		if (ef->fld_type ==  MAILIMF_FIELD_FROM) {
			clistiter *cf;
			for (cf = clist_begin(ef->fld_data.fld_from->frm_mb_list->mb_list);
			    cf; cf = clist_next(cf)) {
				struct mailimf_mailbox *mb = clist_content(cf);
				
				*from = strdup(mb->mb_addr_spec);
			}
		}
	}
	mailimf_fields_free(f);
	
	return 0;
}

int message_parse_attachments(char *msg, size_t length, struct attachment_list **al)
{
	int r;
	struct mailmime *mm, *mmf;
	size_t current_index = 0;
	clistiter * cur;
	
	*al = attachment_list_alloc();

	for (current_index = 0; current_index < length; current_index++) {
		if (current_index == 0 &&
		    !strncasecmp("Content-Type:", msg, 13))
			break;
		if (!strncasecmp("\r\nContent-Type:", msg + current_index, 15)){
			current_index+=2;
			break;
		}
	}
	if (current_index >= length)
		return -1;

		
	msg+= current_index;
	length -= current_index;

	current_index = 0;
	r = mailmime_parse(msg, length, &current_index, &mm);
	if (r != MAILIMF_NO_ERROR)
		return -1;
	mmf = mm;
	
//	printf("parent type: %d\n", mm->mm_parent_type);
//	printf("type: %d\n", mm->mm_type);
//	printf("ct: %s\n", mm->mm_content_type->ct_subtype);

	if (mm->mm_type == MAILMIME_MESSAGE)
		mm = mm->mm_data.mm_message.mm_msg_mime;
	
//	printf("parent type: %d\n", mm->mm_parent_type);
//	printf("type: %d\n", mm->mm_type);
//	printf("ct: %s\n", mm->mm_content_type->ct_subtype);
	
	for(cur = clist_begin(mm->mm_data.mm_multipart.mm_mp_list) ; cur;
	    cur = clist_next(cur)) {
		struct mailmime *part = clist_content(cur);
//		printf("encoded: %d\n", part->mm_body->dt_encoded);
//		printf("type: %d\n", part->mm_type);
		size_t current_index;
		char * result;
		size_t result_len;
	
		current_index = 0;
		r = mailmime_part_parse(
		    part->mm_body->dt_data.dt_text.dt_data,
		    part->mm_body->dt_data.dt_text.dt_length,
		    &current_index,
		    part->mm_body->dt_encoding, 
		    &result, &result_len);
		if (r == MAILIMF_NO_ERROR) {
			struct attachment *a = attachment_alloc();
			clistiter *cur2;
				
			attachment_data_set(a, result, result_len);
				
			attachment_type_set(a,
			    part->mm_content_type->ct_subtype);

			for(cur2 = clist_begin(part->mm_mime_fields->fld_list);
			    cur2 != NULL ;
			    cur2 = clist_next(cur2)) {
				struct mailmime_field *f = clist_content(cur2);
				struct mailmime_disposition *d;
				clistiter *cur3;
				
				if (f->fld_type != MAILMIME_FIELD_DISPOSITION)
					continue;
				d = f->fld_data.fld_disposition;
				
				for (cur3 = clist_begin(d->dsp_parms);
				     cur3 != NULL;
				     cur3 = clist_next(cur3)) {
					struct mailmime_disposition_parm *p;
					p = clist_content(cur3);
					
					if (p->pa_type != MAILMIME_DISPOSITION_PARM_FILENAME)
						continue;
					
					attachment_name_set(a, 
					    p->pa_data.pa_filename);
				}
			}
				    
			attachment_list_add(*al, a);
	  
			mmap_string_unref(result);
		}
		
	}

//	printf("r: %d %zd %zd\n", r, current_index, length);
	
	mailmime_free(mmf);

	return 0;	
}

#define MAX_MESSAGE_ID 512

static char * generate_boundary(const char * boundary_prefix)
{
    char id[MAX_MESSAGE_ID];
    time_t now;
    long value;
    
    now = time(NULL);
    value = random();
    
    if (boundary_prefix == NULL)
        boundary_prefix = "";
    
    snprintf(id, MAX_MESSAGE_ID, "%s%lx_%lx", boundary_prefix, now, value);
    
    return strdup(id);
}


static struct mailmime *part_new_empty(struct mailmime_content * content,
    struct mailmime_fields * mime_fields,
    const char * boundary_prefix,
    int force_single)
{
	struct mailmime * build_info;
	clist * list;
	int r;
	int mime_type;

	list = NULL;

	if (force_single) {
		mime_type = MAILMIME_SINGLE;
	} else {
		switch (content->ct_type->tp_type) {
			case MAILMIME_TYPE_DISCRETE_TYPE:
				mime_type = MAILMIME_SINGLE;
				break;

			case MAILMIME_TYPE_COMPOSITE_TYPE:
				switch (content->ct_type->tp_data.tp_composite_type->ct_type) {
					case MAILMIME_COMPOSITE_TYPE_MULTIPART:
						mime_type = MAILMIME_MULTIPLE;
						break;

					case MAILMIME_COMPOSITE_TYPE_MESSAGE:
						if (strcasecmp(content->ct_subtype, "rfc822") == 0)
							mime_type = MAILMIME_MESSAGE;
						else
							mime_type = MAILMIME_SINGLE;
						break;

					default:
						goto err;
				}
				break;

			default:
				goto err;
		}
	}

	if (mime_type == MAILMIME_MULTIPLE) {
		char * attr_name;
		char * attr_value;
		struct mailmime_parameter * param;
		clist * parameters;
		char * boundary;

		list = clist_new();
		if (list == NULL)
			goto err;

		attr_name = strdup("boundary");
		boundary = generate_boundary(boundary_prefix);
		attr_value = boundary;
		if (attr_name == NULL) {
			free(attr_name);
			goto free_list;
		}

		param = mailmime_parameter_new(attr_name, attr_value);
		if (param == NULL) {
			free(attr_value);
			free(attr_name);
			goto free_list;
		}

		if (content->ct_parameters == NULL) {
			parameters = clist_new();
			if (parameters == NULL) {
				mailmime_parameter_free(param);
				goto free_list;
			}
		} else
			parameters = content->ct_parameters;

		r = clist_append(parameters, param);
		if (r != 0) {
			clist_free(parameters);
			mailmime_parameter_free(param);
			goto free_list;
		}

		if (content->ct_parameters == NULL)
			content->ct_parameters = parameters;
	}

	build_info = mailmime_new(mime_type,
	    NULL, 0, mime_fields, content,
	    NULL, NULL, NULL, list,
	    NULL, NULL);
	if (build_info == NULL) {
		clist_free(list);
		return NULL;
	}

	return build_info;

free_list:
	clist_free(list);
err:
	return NULL;
}

static struct mailmime * part_multiple_new(const char * type, const char * boundary_prefix)
{
    struct mailmime_fields * mime_fields;
    struct mailmime_content * content;
    struct mailmime * mp;
    
    mime_fields = mailmime_fields_new_empty();
    if (mime_fields == NULL)
        goto err;
    
    content = mailmime_content_new_with_str(type);
    if (content == NULL)
        goto free_fields;
    
    mp = part_new_empty(content, mime_fields, boundary_prefix, 0);
    if (mp == NULL)
        goto free_content;
    
    return mp;
    
free_content:
    mailmime_content_free(content);
free_fields:
    mailmime_fields_free(mime_fields);
err:
    return NULL;
}

bool check_7bit(const char *text, size_t length)
{
	size_t i;
	size_t linelen = 0;
	
	for (i = 0; i < length; i++) {
		if (text[i] == 0)
			return false;
		if (text[i] > 127)
			return false;
		if (text[i] == '\n')
			linelen = 0;
		if (linelen > 999)
			return false;
	}
	
	return true;
}

static struct mailmime * get_file_part(const char * filename, const char * mime_type,
                                       const char * text, size_t length)
{
	char * disposition_name;
	int encoding_type;
	struct mailmime_disposition * disposition;
	struct mailmime_mechanism * encoding;
	struct mailmime_content * content;
	struct mailmime * mime;
	struct mailmime_fields * mime_fields;

	disposition_name = NULL;
	if (filename != NULL) {
		disposition_name = strdup(filename);
	}
	disposition = mailmime_disposition_new_with_data(MAILMIME_DISPOSITION_TYPE_ATTACHMENT,
	disposition_name, NULL, NULL, NULL, (size_t) -1);
	content = mailmime_content_new_with_str(mime_type);

	if (check_7bit(text, length))
		encoding_type = MAILMIME_MECHANISM_7BIT;
	else
		encoding_type = MAILMIME_MECHANISM_BASE64;
	encoding = mailmime_mechanism_new(encoding_type, NULL);
	mime_fields = mailmime_fields_new_with_data(encoding,
	    NULL, NULL, disposition, NULL);
	mime = part_new_empty(content, mime_fields, NULL, 1);
	mailmime_set_body_text(mime, (char *) text, length);

	return mime;
}

static struct mailimf_fields * build_fields(char *tostr)
{
	struct mailimf_fields * fields;
	struct mailimf_field * f;
	clist * list;
	struct mailimf_from * from;
	struct mailimf_to * to;
	struct mailimf_mailbox * mb;
	struct mailimf_address * addr;
	struct mailimf_mailbox_list * mb_list;
	struct mailimf_address_list * addr_list;
	struct mailimf_subject *subject;
	clist * fields_list;

	/* build headers */

	fields_list = clist_new();
  
	/* build header 'From' */
  
	list = clist_new();
	mb = mailimf_mailbox_new(NULL, strdup(settings_address_get()));
	clist_append(list, mb);
	mb_list = mailimf_mailbox_list_new(list);
  
	from = mailimf_from_new(mb_list);
  
	f = mailimf_field_new(MAILIMF_FIELD_FROM,
	    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
	    from, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
	    NULL, NULL, NULL, NULL);

	clist_append(fields_list, f);
  
	/* build header To */

	list = clist_new();
	mb = mailimf_mailbox_new(NULL, strdup(tostr));
	addr = mailimf_address_new(MAILIMF_ADDRESS_MAILBOX, mb, NULL);
	clist_append(list, addr);
	addr_list = mailimf_address_list_new(list);
  
	to = mailimf_to_new(addr_list);

	f = mailimf_field_new(MAILIMF_FIELD_TO,
	    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
	    NULL, NULL, NULL, to, NULL, NULL, NULL, NULL, NULL,
	    NULL, NULL, NULL, NULL);
  
	clist_append(fields_list, f);

	f = mailimf_field_new_custom(strdup("X-QuickMSG"), strdup("v1.0"));
	clist_append(fields_list, f);

	subject = mailimf_subject_new(strdup("[QuickMSG]"));

	f = mailimf_field_new(MAILIMF_FIELD_SUBJECT,
	    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
	    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
	    subject, NULL, NULL, NULL);
  
	clist_append(fields_list, f);

	fields = mailimf_fields_new(fields_list);
  
	return fields;
}



int message_multipart_create(char *to, struct attachment_list *l,
    void **msg, size_t *length, char *type)
{
	struct mailmime * msg_mime;
	struct mailmime * mime;
//	struct mailmime * submime;
	struct mailimf_fields * fields;
	size_t i;
	int col;
	FILE *fd;
	char **msgc = (char **)msg;
	
	msg_mime = mailmime_new_message_data(NULL);
	if (to) {
		fields = build_fields(to);
		mailmime_set_imf_fields(msg_mime, fields);
	}

	mime = part_multiple_new(type, NULL);

	for (i = 0; i < l->size; i++) {
		struct mailmime * part;
		char *mimetype;
		
		asprintf(&mimetype, "%s/%s", l->attachments[i]->basetype,
		    l->attachments[i]->type);
		
//		printf("mimetype: %s\n", mimetype);

		part = get_file_part(l->attachments[i]->name, mimetype,
			l->attachments[i]->data, l->attachments[i]->length);

		mailmime_smart_add_part(mime, part);
		
		free(mimetype);
	}

	mailmime_add_part(msg_mime, mime);

	fd = open_memstream(msgc, length);

	col = 0;
	if (to)
		mailmime_write(fd, &col, msg_mime);
	else
		mailmime_write(fd, &col, mime);

	mailmime_free(msg_mime);

	fclose(fd);

	return 0;
}
